"""
CSV file wrapper.
Authors: Jacky Liang, Jeff Mahler
"""
import os, shutil, logging, csv

class CSVModel:
    """A generic model for CSV file reading and writing.
    """

    _KNOWN_TYPES_MAP = {
        'str': str,
        'int': int,
        'float': float,
        'bool': bool
    }

    def __init__(self, full_filename, headers_types_list, default_entry=''):
        """Instantiates a CSVModel object.

        Parameters
        ----------
        full_filename : :obj:`str`
            The file path to a .csv file.

        headers_types_list : :obj:`list` of two-tuples :obj:`str`, :obj:`str`
            A list where each item is a tuple of string header for a column and
            the correspoding data type as a string.

        default_entry : :obj:`str`
            The default entry for cells in the CSV.

        Raises
        ------
        Exception
            If the types, headers, or default entry are not strings.
        """
        headers, types = zip(*headers_types_list)
        headers_types = {headers[i]:types[i] for i in range(len(headers))}
        for key, val in headers_types.items():
            if type(val) != str:
                raise Exception("Types must be passed in as strings! For header {0}, type {1}".format(key, val))
            if val not in CSVModel._KNOWN_TYPES_MAP:
                raise Exception("Invalid type for header {0}. Got: {1}. Can only accept: {2}".format(key, val, CSVModel._KNOWN_TYPES_MAP.keys()))
            if key == '_uid' or key == '_default':
                raise Exception("Cannot create reserved columns _uid or _default!")
        if type(default_entry) != str:
            raise Exception('Default entry must be a string! Got: {0}'.format(default_entry))

        self._headers_types = headers_types.copy()
        self._headers = ('_uid',) + tuple(headers) + ('_default',)

        if not full_filename.endswith('.csv'):
            full_filename = '{0}.csv'.format(full_filename)

        self._full_filename = full_filename
        filename = os.path.basename(full_filename)
        file_path = os.path.dirname(full_filename)
        self._full_backup_filename = os.path.join(file_path, '.backup_{0}'.format(filename))

        self._uid = 0
        self._default_entry = default_entry

        self._table = []

        if not os.path.exists(file_path) and file_path != '':
            os.makedirs(file_path)

        types_row = self._headers_types.copy()
        types_row['_uid'] = 'int'
        types_row['_default'] = self._default_entry
        self._table.append(types_row)
        self._save()

    def _get_new_uid(self):
        """Get a new UID by incrementing the cached UID.

        Returns
        -------
        int
            The new UID.
        """
        return_uid = self._uid
        self._uid += 1
        return return_uid

    @property
    def num_rows(self):
        """int : The number of rows in the table.
        """
        return len(self._table) - 1

    def get_cur_uid(self):
        """Get the current UID associated with this object.

        Returns
        -------
        int
            The new UID.
        """
        return self._uid

    def _save(self):
        """Save the model to a .csv file
        """
        # if not first time saving, copy .csv to a backup
        if os.path.isfile(self._full_filename):
            shutil.copyfile(self._full_filename, self._full_backup_filename)

        # write to csv
        with open(self._full_filename, 'w') as file:
            writer = csv.DictWriter(file, fieldnames=self._headers)
            writer.writeheader()
            for row in self._table:
                writer.writerow(row)

    def insert(self, data):
        """Insert a row into the .csv file.

        Parameters
        ----------
        data : :obj:`dict`
            A dictionary mapping keys (header strings) to values.

        Returns
        -------
        int
            The UID for the new row.

        Raises
        ------
        Exception
            If the value for a given header is not of the appropriate type.
        """
        row = {key:self._default_entry for key in self._headers}
        row['_uid'] = self._get_new_uid()

        for key, val in data.items():
            if key in ('_uid', '_default'):
                logging.warn("Cannot manually set columns _uid or _default of a row! Given data: {0}".format(data))
                continue
            if not isinstance(val, CSVModel._KNOWN_TYPES_MAP[self._headers_types[key]]):
                raise Exception('Data type mismatch for column {0}. Expected: {1}, got: {2}'.format(key,
                                                        CSVModel._KNOWN_TYPES_MAP[self._headers_types[key]], type(val)))
            row[key] = val

        self._table.append(row)
        self._save()
        return row['_uid']

    def update_by_uid(self, uid, data):
        """Update a row with the given data.

        Parameters
        ----------
        uid : int
            The UID of the row to update.

        data : :obj:`dict`
            A dictionary mapping keys (header strings) to values.

        Raises
        ------
        Exception
            If the value for a given header is not of the appropriate type.
        """
        row = self._table[uid+1]
        for key, val in data.items():
            if key == '_uid' or key == '_default':
                continue
            if key not in self._headers:
                logging.warn("Unknown column name: {0}".format(key))
                continue
            if not isinstance(val, CSVModel._KNOWN_TYPES_MAP[self._headers_types[key]]):
                raise Exception('Data type mismatch for column {0}. Expected: {1}, got: {2}'.format(key,
                                                        CSVModel._KNOWN_TYPES_MAP[self._headers_types[key]], type(val)))
            row[key] = val
        self._save()

    def get_by_uid(self, uid):
        """Get a copy of a given row of the table.

        Parameters
        ----------
        uid : int
            The UID of the row to update.

        Returns
        -------
        :obj:`dict`
            A dictionary mapping keys (header strings) to values, which
            represents a row of the table.
        """
        return self._table[uid+1].copy()

    def get_by_row(self, row):
        """Get a copy of a given row of the table.

        Parameters
        ----------
        row : int
            The number of the row to update.

        Returns
        -------
        :obj:`dict`
            A dictionary mapping keys (header strings) to values, which
            represents a row of the table.
        """
        return self._table[row + 1].copy()

    def get_col(self, col_name, filter = lambda _ : True):
        """Return all values in the column corresponding to col_name that satisfies filter, which is
        a function that takes in a value of the column's type and returns True or False

        Parameters
        ----------
        col_name : str
            Name of desired column
        filter : function, optional
            A function that takes in a value of the column's type and returns True or False
            Defaults to a function that always returns True

        Returns
        -------
        list
            A list of values in the desired columns by order of their storage in the model

        Raises
        ------
        ValueError
            If the desired column name is not found in the model
        """
        if col_name not in self._headers:
            raise ValueError("{} not found! Model has headers: {}".format(col_name, self._headers))
        col = []
        for i in range(self.num_rows):
            row = self._table[i + 1]
            val = row[col_name]
            if filter(val):
                col.append(val)

        return col

    def get_by_cols(self, cols, direction=1):
        """Return the first or last row that satisfies the given col value constraints,
        or None if no row contains the given value.

        Parameters
        ----------
        cols: :obj:'dict'
            Dictionary of col values for a specific row.
        direction: int, optional
            Either 1 or -1. 1 means find the first row, -1 means find the last row.

        Returns
        -------
        :obj:`dict`
            A dictionary mapping keys (header strings) to values, which
            represents a row of the table. This row contains the given value in
            the specified column.
        """
        if direction == 1:
            iterator = range(self.num_rows)
        elif direction == -1:
            iterator = range(self.num_rows-1, -1, -1)
        else:
            raise ValueError("Direction can only be 1 (first) or -1 (last). Got: {0}".format(direction))

        for i in iterator:
            row = self._table[i+1]

            all_sat = True
            for key, val in cols.items():
                if row[key] != val:
                    all_sat = False
                    break

            if all_sat:
                return row.copy()

        return None

    def get_by_col(self, col, val):
        """Return the first row that contains the given value in the specified column,
        or None if no row contains the given value.

        Parameters
        ----------
        col : :obj:`str`
            The header string for a column.

        val : value type
            The value to match in the column.

        Returns
        -------
        :obj:`dict`
            A dictionary mapping keys (header strings) to values, which
            represents a row of the table. This row contains the given value in
            the specified column.
        """
        return self.get_by_cols({col:val}, direction=1)

    def get_by_col_last(self, col, val):
        """Return the last row that contains the given value in the specified column,
        or None if no row contains the given value.

        Parameters
        ----------
        col : :obj:`str`
            The header string for a column.

        val : value type
            The value to match in the column.

        Returns
        -------
        :obj:`dict`
            A dictionary mapping keys (header strings) to values, which
            represents a row of the table. This row contains the given value in
            the specified column.
        """
        return self.get_by_cols({col:val}, direction=-1)

    def get_rows_by_cols(self, matching_dict):
        """Return all rows where the cols match the elements given in the matching_dict

        Parameters
        ----------
        matching_dict: :obj:'dict'
            Desired dictionary of col values.

        Returns
        -------
        :obj:`list`
            A list of rows that satisfy the matching_dict
        """
        result = []
        for i in range(self.num_rows):
            row = self._table[i+1]
            matching = True
            for key, val in matching_dict.items():
                if row[key] != val:
                    matching = False
                    break

            if matching:
                result.append(row)

        return result

    def __iter__(self):
        """ Forms an iterator """
        self._cur_row = 1
        return self

    def next(self):
        """ Returns the next row in the CSV, for iteration """
        if self._cur_row >= len(self._table):
            raise StopIteration
        data = self._table[self._cur_row].copy()
        self._cur_row += 1
        return data

    @staticmethod
    def _str_to_bool(s):
        truthy = {'T', 'True', 't', 'true', 'TRUE'}
        falsy = {'F', 'False', 'f', 'false', 'FALSE'}

        if s in truthy:
            return True
        elif s in falsy:
            return False
        else:
            raise ValueError('Cannot convert {} to a boolean value! Accepted values are {} and {}'.format(s, truthy, falsy))

    @staticmethod
    def load(full_filename):
        """Load a .csv file into a CSVModel.

        Parameters
        ----------
        full_filename : :obj:`str`
            The file path to a .csv file.

        Returns
        -------
        :obj:`CSVModel`
            The CSVModel initialized with the data in the given file.

        Raises
        ------
        Excetpion
            If the CSV file does not exist or is malformed.
        """
        with open(full_filename, 'r') as file:
            reader = csv.DictReader(file)
            headers = reader.fieldnames
            if '_uid' not in headers or '_default' not in headers:
                raise Exception("Malformed CSVModel file!")

            all_rows = [row for row in reader]
            types = all_rows[0]
            table = [types]
            default_entry = table[0]['_default']

            for i in range(1, len(all_rows)):
                raw_row = all_rows[i]
                row = {}
                for column_name in headers:
                    if raw_row[column_name] != default_entry and column_name != '':
                        if types[column_name] == 'bool':
                            row[column_name] = CSVModel._str_to_bool(raw_row[column_name])
                        else:
                            try:
                                row[column_name] = CSVModel._KNOWN_TYPES_MAP[types[column_name]](raw_row[column_name])
                            except:
                                logging.error('{}, {}, {}'.format(column_name, types[column_name], raw_row[column_name]))
                                row[column_name] = CSVModel._KNOWN_TYPES_MAP[types[column_name]](bool(raw_row[column_name]))
                    else:
                        row[column_name] = default_entry
                table.append(row)

        if len(table) == 1:
            next_valid_uid = 0
        else:
            next_valid_uid = int(table[-1]['_uid']) + 1

        headers_init = headers[1:-1]
        types_init = [types[column_name] for column_name in headers_init]
        headers_types_list = zip(headers_init, types_init)

        csv_model = CSVModel(full_filename, headers_types_list, default_entry=default_entry)
        csv_model._uid = next_valid_uid
        csv_model._table = table
        csv_model._save()
        return csv_model

    @staticmethod
    def get_or_create(full_filename, headers_types=None, default_entry=''):
        """Load a .csv file into a CSVModel if the file exists, or create
        a new CSVModel with the given filename if the file does not exist.

        Parameters
        ----------
        full_filename : :obj:`str`
            The file path to a .csv file.

        headers_types : :obj:`list` of :obj:`tuple` of :obj:`str`, :obj:`str`
            A list of tuples, where the first element in each tuple is the
            string header for a column and the second element is that column's
            data type as a string.

        default_entry : :obj:`str`
            The default entry for cells in the CSV.

        Returns
        -------
        :obj:`CSVModel`
            The CSVModel initialized with the data in the given file, or a new
            CSVModel tied to the filename if the file doesn't currently exist.
        """
        # convert dictionaries to list
        if isinstance(headers_types, dict):
            headers_types_list = [(k,v) for k,v in headers_types.items()]
            headers_types = headers_types_list

        if os.path.isfile(full_filename):
            return CSVModel.load(full_filename)
        else:
            return CSVModel(full_filename, headers_types, default_entry=default_entry)



# -*- coding: utf-8 -*-
"""
Copyright ©2017. The Regents of the University of California (Regents). All Rights Reserved.
Permission to use, copy, modify, and distribute this software and its documentation for educational,
research, and not-for-profit purposes, without fee and without a signed licensing agreement, is
hereby granted, provided that the above copyright notice, this paragraph and the following two
paragraphs appear in all copies, modifications, and distributions. Contact The Office of Technology
Licensing, UC Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-
7201, otl@berkeley.edu, http://ipira.berkeley.edu/industry-info for commercial licensing opportunities.

IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED
HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO OBLIGATION TO PROVIDE
MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
"""
# Access levels
READ_ONLY_ACCESS = 'READ_ONLY'
READ_WRITE_ACCESS = 'READ_WRITE'
WRITE_ACCESS = 'WRITE'

# Formatting
JSON_INDENT = 2

# Train / Test
TRAIN_ID = 0
TEST_ID = 1



"""
Commonly used helper functions
Author: Jeff Mahler
"""
import logging
import os
from six.moves import input

import numpy as np

def gen_experiment_id(n=10):
    """Generate a random string with n characters.

    Parameters
    ----------
    n : int
        The length of the string to be generated.

    Returns
    -------
    :obj:`str`
        A string with only alphabetic characters. 
    """
    chrs = 'abcdefghijklmnopqrstuvwxyz'
    inds = np.random.randint(0,len(chrs), size=n)
    return ''.join([chrs[i] for i in inds])

def get_elapsed_time(time_in_seconds):
    """ Helper function to get elapsed time in human-readable format.

    Parameters
    ----------
    time_in_seconds : float
        runtime, in seconds

    Returns
    -------
    str
        formatted human-readable string describing the time
    """
    if time_in_seconds < 60:
        return '%.1f seconds' % (time_in_seconds)
    elif time_in_seconds < 3600:
        return '%.1f minutes' % (time_in_seconds / 60)
    else:
        return '%.1f hours' % (time_in_seconds / 3600)

def mkdir_safe(path):
    """ Creates a directory if it does not already exist.

    Parameters
    ----------
    path : str
        path to the directory to create

    Returns
    -------
    bool
        True if the directory was created, False otherwise
    """
    if not os.path.exists(path):
        os.mkdir(path)
    
def histogram(values, num_bins, bounds, normalized=True, plot=False, color='b'):
    """Generate a histogram plot.

    Parameters
    ----------
    values : :obj:`numpy.ndarray`
        An array of values to put in the histogram.

    num_bins : int
        The number equal-width bins in the histogram.

    bounds : :obj:`tuple` of float
        Two floats - a min and a max - that define the lower and upper
        ranges of the histogram, respectively.

    normalized : bool
        If True, the bins will show the percentage of elements they contain
        rather than raw counts.

    plot : bool
        If True, this function uses pyplot to plot the histogram.

    color : :obj:`str`
        The color identifier for the plotted bins.

    Returns
    -------
    :obj:`tuple of `:obj:`numpy.ndarray`
        The values of the histogram and the bin edges as ndarrays.
    """
    hist, bins = np.histogram(values, bins=num_bins, range=bounds)
    width = (bins[1] - bins[0])
    if normalized:
        if np.sum(hist) > 0:
            hist = hist.astype(np.float32) / np.sum(hist)
    if plot:
        import matplotlib.pyplot as plt
        plt.bar(bins[:-1], hist, width=width, color=color)    
    return hist, bins

def skew(xi):
    """Return the skew-symmetric matrix that can be used to calculate
    cross-products with vector xi.

    Multiplying this matrix by a vector `v` gives the same result
    as `xi x v`.

    Parameters
    ----------
    xi : :obj:`numpy.ndarray` of float
        A 3-entry vector.

    Returns
    -------
    :obj:`numpy.ndarray` of float
        The 3x3 skew-symmetric cross product matrix for the vector.
    """
    S = np.array([[0, -xi[2], xi[1]],
                  [xi[2], 0, -xi[0]],
                  [-xi[1], xi[0], 0]])
    return S

def deskew(S):
    """Converts a skew-symmetric cross-product matrix to its corresponding
    vector. Only works for 3x3 matrices.

    Parameters
    ----------
    S : :obj:`numpy.ndarray` of float
        A 3x3 skew-symmetric matrix.

    Returns
    -------
    :obj:`numpy.ndarray` of float
        A 3-entry vector that corresponds to the given cross product matrix.
    """
    x = np.zeros(3)
    x[0] = S[2,1]
    x[1] = S[0,2]
    x[2] = S[1,0]
    return x

def reverse_dictionary(d):
    """ Reverses the key value pairs for a given dictionary.

    Parameters
    ----------
    d : :obj:`dict`
        dictionary to reverse

    Returns
    -------
    :obj:`dict`
        dictionary with keys and values swapped
    """
    rev_d = {}
    [rev_d.update({v:k}) for k, v in d.items()]
    return rev_d

def pretty_str_time(dt):
    """Get a pretty string for the given datetime object.
    
    Parameters
    ----------
    dt : :obj:`datetime`
        A datetime object to format.
    
    Returns
    -------
    :obj:`str`
        The `datetime` formatted as {year}_{month}_{day}_{hour}_{minute}.
    """
    return "{0}_{1}_{2}_{3}:{4}".format(dt.year, dt.month, dt.day, dt.hour, dt.minute)

def filenames(directory, tag='', sorted=False):
    """ Reads in all filenames from a directory that contain a specified substring.

    Parameters
    ----------
    directory : :obj:`str`
        the directory to read from
    tag : :obj:`str`
        optional tag to match in the filenames
    sorted : bool
        whether or not to sort the filenames

    Returns
    -------
    :obj:`list` of :obj:`str`
        filenames to read from
    """
    f = [os.path.join(directory, f) for f in os.listdir(directory) if f.find(tag) > -1]
    if sorted:
        f.sort()
    return f

def sph2cart(r, az, elev):
    """ Convert spherical to cartesian coordinates.

    Attributes
    ----------
    r : float
        radius
    az : float
        aziumth (angle about z axis)
    elev : float
        elevation from xy plane

    Returns
    -------
    float
        x-coordinate
    float
        y-coordinate
    float
        z-coordinate
    """
    x = r * np.cos(az) * np.sin(elev)
    y = r * np.sin(az) * np.sin(elev)
    z = r * np.cos(elev)
    return x, y, z

def cart2sph(x, y, z):
    """ Convert cartesian to spherical coordinates.

    Attributes
    ----------
    x : float
        x-coordinate
    y : float
        y-coordinate
    z : float
        z-coordinate

    Returns
    -------
    float
        radius
    float
        aziumth
    float
        elevation
    """
    r = np.sqrt(x**2 + y**2 + z**2)
    if x > 0 and y > 0:
        az = np.arctan(y / x)
    elif x > 0 and y < 0:
        az = 2*np.pi - np.arctan(-y / x)
    elif x < 0 and y > 0:
        az = np.pi - np.arctan(-y / x)    
    elif x < 0 and y < 0:
        az = np.pi + np.arctan(y / x)    
    elif x == 0 and y > 0:
        az = np.pi / 2
    elif x == 0 and y < 0:
        az = 3 * np.pi / 2
    elif y == 0 and x > 0:
        az = 0
    elif y == 0 and x < 0:
        az = np.pi
    elev = np.arccos(z / r)
    return r, az, elev

def keyboard_input(message, yesno=False):
    """ Get keyboard input from a human, optionally reasking for valid
    yes or no input.

    Parameters
    ----------
    message : :obj:`str`
        the message to display to the user
    yesno : :obj:`bool`
        whether or not to enforce yes or no inputs
    
    Returns
    -------
    :obj:`str`
        string input by the human
    """
    # add space for readability
    message += ' '

    # add yes or no to message
    if yesno:
        message += '[y/n] '

    # ask human
    human_input = input(message)
    if yesno:
        while human_input.lower() != 'n' and human_input.lower() != 'y':
            logging.info('Did not understand input. Please answer \'y\' or \'n\'')
            human_input = input(message)
    return human_input

def sqrt_ceil(n):
    """ Computes the square root of an number rounded up to the nearest integer. Very useful for plotting.

    Parameters
    ----------
    n : int
        number to sqrt

    Returns
    -------
    int
        the sqrt rounded up to the nearest integer
    """
    return int(np.ceil(np.sqrt(n)))



# -*- coding: utf-8 -*-
"""
Copyright ©2017. The Regents of the University of California (Regents). All Rights Reserved.
Permission to use, copy, modify, and distribute this software and its documentation for educational,
research, and not-for-profit purposes, without fee and without a signed licensing agreement, is
hereby granted, provided that the above copyright notice, this paragraph and the following two
paragraphs appear in all copies, modifications, and distributions. Contact The Office of Technology
Licensing, UC Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-
7201, otl@berkeley.edu, http://ipira.berkeley.edu/industry-info for commercial licensing opportunities.

IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED
HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO OBLIGATION TO PROVIDE
MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
"""
"""
Helper classes for analyzing machine learning output
Author: Jeff Mahler
"""
import numpy as np
import os
import ruamel.yaml as yaml

import sklearn.metrics as sm
import scipy.stats as ss

class ConfusionMatrix(object):
    """ Confusion matrix for classification errors """
    def __init__(self, num_categories):
        self.num_categories = num_categories

        # organized as row = true, column = pred
        self.data = np.zeros([num_categories, num_categories])

    def update(self, predictions, labels):
        num_pred = predictions.shape[0]
        for i in range(num_pred):
            self.data[labels[i].astype(np.uint16), predictions[i].astype(np.uint16)] += 1

class ClassificationResult(object):
    def __init__(self, pred_probs, labels):
        """ Creates a classification result.

        Parameters
        ----------
        pred_probs : :obj:`numpy.ndarray`
            array of predicted class probabilities
        labels : :obj:`numpy.ndarray`
            array of integer class labels
        """
        if not isinstance(pred_probs, np.ndarray):
            pred_probs = np.array(pred_probs)
        if not isinstance(labels, np.ndarray):
            labels = np.array(labels)        
        self.pred_probs = pred_probs.astype(np.float32)
        self.labels = labels.astype(np.uint32)

    @property
    def error_rate(self):
        return 100.0 - (
            100.0 *
            np.sum(self.predictions == self.labels) /
            self.num_datapoints)

    @property
    def accuracy(self):
        return np.mean(1 * (self.predictions == self.labels))

    def top_k_error_rate(self, k):
        predictions_arr = self.top_k_predictions(k)
        labels_arr = np.zeros(predictions_arr.shape)
        for i in range(k):
            labels_arr[:,i] = self.labels

        return 100.0 - (
            100.0 *
            np.sum(predictions_arr == labels_arr) /
            self.num_datapoints)

    @property
    def fpr(self):
        if np.sum(self.labels == 0) == 0:
            return 0.0
        return float(np.sum((self.predictions == 1) & (self.labels == 0))) / np.sum(self.labels == 0)

    @property
    def precision(self):
        if np.sum(self.predictions == 1) == 0:
            return 1.0
        return float(np.sum((self.predictions == 1) & (self.labels == 1))) / np.sum(self.predictions == 1)

    @property
    def recall(self):
        if np.sum(self.predictions == 1) == 0:
            return 1.0
        return float(np.sum((self.predictions == 1) & (self.labels == 1))) / np.sum(self.labels == 1)

    @property
    def num_datapoints(self):
        return self.pred_probs.shape[0]

    @property
    def num_categories(self):
        return self.pred_probs.shape[1]
        
    @property
    def predictions(self):
        return np.argmax(self.pred_probs, 1)

    def top_k_predictions(self, k):
        return np.argpartition(self.pred_probs, -k, axis=1)[:, -k:]

    @property
    def confusion_matrix(self):
        cm = ConfusionMatrix(self.num_categories)
        cm.update(self.predictions, self.labels)
        return cm

    def mispredicted_indices(self):
        return np.where(self.predictions != self.labels)[0]

    def correct_indices(self):
        return np.where(self.predictions == self.labels)[0]

    def convert_labels(self, mapping):
        new_num_categories = len(set(mapping.values()))
        new_probs = np.zeros([self.num_datapoints, new_num_categories])
        new_labels = np.zeros(self.num_datapoints)
        for i in range(self.num_datapoints):
            for j in range(self.num_categories):
                new_probs[i,mapping[j]] += self.pred_probs[i,j]
            new_labels[i] = mapping[self.labels[i]]
        return ClassificationResult([new_probs], [new_labels])

    @property
    def label_vectors(self):
        label_mat = np.zeros(self.pred_probs.shape)
        for i in range(self.num_datapoints):
            label_mat[i, self.labels[i]] = 1

        pred_probs_vec = self.pred_probs.ravel()
        labels_vec = label_mat.ravel()
        return pred_probs_vec, labels_vec

    def precision_recall_curve(self, plot=False, line_width=2, font_size=15, color='b', style='-', label='', marker=None):
        pred_probs_vec, labels_vec = self.label_vectors
        precision, recall, thresholds = sm.precision_recall_curve(labels_vec, pred_probs_vec)
        if plot:
            import matplotlib.pyplot as plt
            plt.plot(recall, precision, linewidth=line_width, color=color, linestyle=style, label=label, marker=marker)
            plt.xlim(0,1)
            plt.ylim(0,1)
            plt.xlabel('Recall', fontsize=font_size)
            plt.ylabel('Precision', fontsize=font_size)
        return precision, recall, thresholds

    def roc_curve(self, plot=False, line_width=2, font_size=15, color='b', style='-', label=''):
        pred_probs_vec, labels_vec = self.label_vectors
        fpr, tpr, thresholds = sm.roc_curve(labels_vec, pred_probs_vec)

        if plot:
            import matplotlib.pyplot as plt
            plt.plot(fpr, tpr, linewidth=line_width, color=color, linestyle=style, label=label)
            plt.xlabel('FPR', fontsize=font_size)
            plt.ylabel('TPR', fontsize=font_size)
        return fpr, tpr, thresholds

    @property
    def ap_score(self):
        pred_probs_vec, labels_vec = self.label_vectors
        ap = 0.0
        try:
            ap = sm.average_precision_score(labels_vec, pred_probs_vec)
        except:
            pass
        return ap

    @property
    def auc_score(self):
        pred_probs_vec, labels_vec = self.label_vectors
        auc = 0.0
        try:
            auc = sm.roc_auc_score(labels_vec, pred_probs_vec)
        except:
            pass
        return auc
            
    @property
    def pearson_correlation(self):
        pred_probs_vec, labels_vec = self.label_vectors
        pcorr_coef = np.corrcoef(labels_vec, pred_probs_vec)
        pearson_coef = pcorr_coef[0,1]
        return pearson_coef

    @property
    def spearman_correlation(self):
        pred_probs_vec, labels_vec = self.label_vectors
        spearman_coef, _ = ss.spearmanr(labels_vec, pred_probs_vec)
        return spearman_coef

    @property
    def spearman_pvalue(self):
        pred_probs_vec, labels_vec = self.label_vectors
        _, p_value = ss.spearmanr(labels_vec, pred_probs_vec)
        return p_value

    def save(self, filename):
        if not os.path.exists(filename):
            os.mkdir(filename)
        
        pred_filename = os.path.join(filename, 'predictions.npz')
        np.savez_compressed(pred_filename, self.pred_probs)

        labels_filename = os.path.join(filename, 'labels.npz')
        np.savez_compressed(labels_filename, self.labels)

    @staticmethod
    def load(filename):
        if not os.path.exists(filename):
            raise ValueError('File %s does not exists' %(filename))

        pred_filename = os.path.join(filename, 'predictions.npz')
        pred_probs = np.load(pred_filename)['arr_0']

        labels_filename = os.path.join(filename, 'labels.npz')
        labels = np.load(labels_filename)['arr_0']
        return ClassificationResult(pred_probs, labels)

    @staticmethod
    def make_summary_table(train_result, val_result, plot=True, save_dir=None, prepend="", save=False):
        """
        Makes a matplotlib table object with relevant data.
        Thanks to Lucas Manuelli for the contribution.

        Parameters
        ----------
        train_result: ClassificationResult
            result on train split

        val_result: ClassificationResult
            result on validation split

        save_dir: str
            path pointing to where to save results

        Returns
        -------
        dict
            dict with stored values, can be saved to a yaml file

        :obj:`matplotlibt.pyplot.fig`
            a figure containing the table
        """
        table_key_list = ['error_rate', 'recall_at_99_precision', 'average_precision', 'precision', 'recall']
        num_fields = len(table_key_list)
        
        import matplotlib.pyplot as plt
        ax = plt.subplot(111, frame_on=False)
        ax.xaxis.set_visible(False)
        ax.yaxis.set_visible(False)
        
        data = np.zeros([num_fields, 2])
        data_dict = dict()

        names = ['train', 'validation']
        for name, result in zip(names, [train_result, val_result]):
            data_dict[name] = {}
            data_dict[name]['error_rate'] = result.error_rate
            data_dict[name]['average_precision'] = result.ap_score * 100
            data_dict[name]['precision'] = result.precision * 100
            data_dict[name]['recall'] = result.recall * 100

            precision_array, recall_array, _ = result.precision_recall_curve()
            recall_at_99_precision = recall_array[np.argmax(precision_array > 0.99)] * 100  # to put it in percentage terms
            data_dict[name]['recall_at_99_precision'] = recall_at_99_precision
                        
            for i, key in enumerate(table_key_list):
                data_dict[name][key] = float("{0:.2f}".format(data_dict[name][key]))
                j = names.index(name)
                data[i, j] = data_dict[name][key]

        table = plt.table(cellText=data, rowLabels=table_key_list, colLabels=names)

        fig = plt.gcf()
        fig.subplots_adjust(bottom=0.15)
        
        if plot:
            plt.show()

        # save the results
        if save_dir is not None and save:
            fig_filename = os.path.join(save_dir, prepend + 'summary.png')
            yaml_filename = os.path.join(save_dir, prepend + 'summary.yaml')
            
            yaml.dump(data_dict, open(yaml_filename, 'w'), default_flow_style=False)
            fig.savefig(fig_filename, bbox_inches="tight")
            
        return data_dict, fig
    
class RegressionResult(object):
    def __init__(self, predictions, labels):
        """ Creates a classification result.
        NOTE: Does not yet work with multidimensional data.

        Parameters
        ----------
        predictions : :obj:`numpy.ndarray`
            array of predicted values
        labels : :obj:`numpy.ndarray`
            array of true values
        """
        self.predictions = predictions
        self.labels = labels
        
    @property
    def mse(self):
        return np.sum((self.predictions - self.labels)**2) / self.num_datapoints

    @property
    def num_datapoints(self):
        return self.predictions.shape[0]

    def save(self, filename):
        if not os.path.exists(filename):
            os.mkdir(filename)
        
        pred_filename = os.path.join(filename, 'predictions.npz')
        np.savez_compressed(pred_filename, self.predictions)

        labels_filename = os.path.join(filename, 'labels.npz')
        np.savez_compressed(labels_filename, self.labels)

    @staticmethod
    def load(filename):
        if not os.path.exists(filename):
            raise ValueError('File %s does not exists' %(filename))

        pred_filename = os.path.join(filename, 'predictions.npz')
        predictions = np.load(pred_filename)['arr_0']

        labels_filename = os.path.join(filename, 'labels.npz')
        labels = np.load(labels_filename)['arr_0']
        return RegressionResult(predictions, labels)

class BinaryClassificationResult(ClassificationResult):
    def __init__(self, pred_probs, labels, threshold=0.5):
        self.threshold = threshold
        ClassificationResult.__init__(self, pred_probs, labels)

    @property
    def num_categories(self):
        return 2

    @property
    def label_vectors(self):
        return self.pred_probs, self.labels

    def top_k_predictions(self, k):
        raise NotImplementedError()

    def top_k_error_rate(self, k):
        raise NotImplementedError()

    @property
    def predictions(self):
        return 1 * (self.pred_probs >= self.threshold)

    @property
    def precision(self):
        if sum(self.predictions) == 0:
            return 1.0
        return sm.precision_score(self.labels, self.predictions)

    @property
    def recall(self):
        return sm.recall_score(self.labels, self.predictions)

    @property
    def tpr(self):
        return self.recall

    @property
    def fpr(self):
        fp = np.sum((self.labels == 0) & (self.predictions == 1))
        an = np.sum(self.labels == 0)
        return float(fp) / an

    @property
    def f1_score(self):
        return sm.f1_score(self.labels, self.predictions)

    @property
    def phi_coef(self):
        return sm.matthews_corrcoef(self.labels, self.predictions)

    @property
    def true_positive_indices(self):
        return np.where((self.labels == 1) & (self.predictions == 1))[0]

    @property
    def false_positive_indices(self):
        return np.where((self.labels == 0) & (self.predictions == 1))[0]

    @property
    def true_negative_indices(self):
        return np.where((self.labels == 0) & (self.predictions == 0))[0]

    @property
    def false_negative_indices(self):
        return np.where((self.labels == 1) & (self.predictions == 0))[0]
    
    @property
    def num_true_pos(self):
        return np.sum(self.labels)

    @property
    def num_true_neg(self):
        return self.num_datapoints - self.num_true_pos

    @property
    def num_false_pos(self):
        return self.false_positive_indices.shape[0]

    @property
    def num_false_neg(self):
        return self.false_negative_indices.shape[0]
    
    @property
    def pct_true_pos(self):
        return np.mean(self.labels)

    @property
    def pct_true_neg(self):
        return 1.0 - self.pct_true_pos

    @property
    def pct_pred_pos(self):
        return np.mean(1 * (self.predictions == 1))

    @property
    def pct_pred_neg(self):
        return 1.0 - self.pct_pred_pos

    @property
    def cross_entropy_loss(self):
        probs = self.pred_probs.copy()
        probs[probs == 1.0] = 0.99999
        probs[probs == 0.0] = 0.00001
        return -np.mean(self.labels * np.log(probs) + (1.0 - self.labels) * np.log(1.0 - probs))
    
    @property
    def sorted_values(self):
        # sort by prob
        labels_and_probs = zip(self.labels, self.pred_probs)
        labels_and_probs.sort(key = lambda x: x[1])
        labels = [l[0] for l in labels_and_probs]
        probs = [l[1] for l in labels_and_probs]
        return labels, probs

    @property
    def app_score(self):
        """ Computes the area under the app curve. """
        # compute curve
        precisions, pct_pred_pos, taus = self.precision_pct_pred_pos_curve(interval=False)

        # compute area
        app = 0
        total = 0
        for k in range(len(precisions)-1):
            # read cur data
            cur_prec = precisions[k]
            cur_pp = pct_pred_pos[k]
            cur_tau = taus[k]

            # read next data
            next_prec = precisions[k+1]
            next_pp = pct_pred_pos[k+1]
            next_tau = taus[k+1]
            
            # approximate with rectangles
            mid_prec = (cur_prec + next_prec) / 2.0
            width_pp = np.abs(next_pp - cur_pp)
            app += mid_prec * width_pp
            total += width_pp

        return app

    def precision_recall_curve(self, plot=False, line_width=2, font_size=15, color='b', style='-', label='', marker=None):
        precision, recall, thresholds = sm.precision_recall_curve(self.labels, self.pred_probs)
        if plot:
            import matplotlib.pyplot as plt
            plt.plot(recall, precision, linewidth=line_width, color=color, linestyle=style, label=label, marker=marker)
            plt.xlim(0,1)
            plt.ylim(0,1)
            plt.xlabel('Recall', fontsize=font_size)
            plt.ylabel('Precision', fontsize=font_size)
        return precision, recall, thresholds

    def roc_curve(self, plot=False, line_width=2, font_size=15, color='b', style='-', label=''):
        fpr, tpr, thresholds = sm.roc_curve(self.labels, self.pred_probs)

        if plot:
            import matplotlib.pyplot as plt
            plt.plot(fpr, tpr, linewidth=line_width, color=color, linestyle=style, label=label)
            plt.xlabel('FPR', fontsize=font_size)
            plt.ylabel('TPR', fontsize=font_size)
        return fpr, tpr, thresholds
    
    def accuracy_curve(self, delta_tau=0.01):
        """ Computes the relationship between probability threshold
        and classification accuracy. """
        # compute thresholds based on the sorted probabilities
        orig_thresh = self.threshold
        sorted_labels, sorted_probs = self.sorted_values

        scores = []
        taus = []
        tau = 0
        for k in range(len(sorted_labels)):
            # compute new accuracy
            self.threshold = tau
            scores.append(self.accuracy)
            taus.append(tau)

            # update threshold
            tau = sorted_probs[k]

        # add last datapoint
        tau = 1.0
        self.threshold = tau
        scores.append(self.accuracy)
        taus.append(tau)

        self.threshold = orig_thresh
        return scores, taus

    def precision_curve(self, delta_tau=0.01):
        """ Computes the relationship between probability threshold
        and classification precision. """
        # compute thresholds based on the sorted probabilities
        orig_thresh = self.threshold
        sorted_labels, sorted_probs = self.sorted_values

        scores = []
        taus = []
        tau = 0
        for k in range(len(sorted_labels)):
            # compute new accuracy
            self.threshold = tau
            scores.append(self.precision)
            taus.append(tau)

            # update threshold
            tau = sorted_probs[k]

        # add last datapoint
        tau = 1.0
        self.threshold = tau
        scores.append(1.0)
        taus.append(tau)

        self.threshold = orig_thresh
        return scores, taus

    def recall_curve(self, delta_tau=0.01):
        """ Computes the relationship between probability threshold
        and classification precision. """
        # compute thresholds based on the sorted probabilities
        orig_thresh = self.threshold
        sorted_labels, sorted_probs = self.sorted_values

        scores = []
        taus = []
        tau = 0
        for k in range(len(sorted_labels)):
            # compute new accuracy
            self.threshold = tau
            scores.append(self.recall)
            taus.append(tau)

            # update threshold
            tau = sorted_probs[k]

        # add last datapoint
        tau = 1.0
        self.threshold = tau
        scores.append(1.0)
        taus.append(tau)

        self.threshold = orig_thresh
        return scores, taus

    def f1_curve(self, delta_tau=0.01):
        """ Computes the relationship between probability threshold
        and classification F1 score. """
        # compute thresholds based on the sorted probabilities
        orig_thresh = self.threshold
        sorted_labels, sorted_probs = self.sorted_values

        scores = []
        taus = []
        tau = 0
        for k in range(len(sorted_labels)):
            # compute new accuracy
            self.threshold = tau
            scores.append(self.f1_score)
            taus.append(tau)

            # update threshold
            tau = sorted_probs[k]

        # add last datapoint
        tau = 1.0
        self.threshold = tau
        scores.append(self.f1_score)
        taus.append(tau)

        self.threshold = orig_thresh
        return scores, taus

    def phi_coef_curve(self, delta_tau=0.01):
        """ Computes the relationship between probability threshold
        and classification phi coefficient. """
        # compute thresholds based on the sorted probabilities
        orig_thresh = self.threshold
        sorted_labels, sorted_probs = self.sorted_values

        scores = []
        taus = []
        tau = 0
        for k in range(len(sorted_labels)):
            # compute new accuracy
            self.threshold = tau
            scores.append(self.phi_coef)
            taus.append(tau)

            # update threshold
            tau = sorted_probs[k]

        # add last datapoint
        tau = 1.0
        self.threshold = tau
        scores.append(self.phi_coef)
        taus.append(tau)

        self.threshold = orig_thresh
        return scores, taus

    def precision_pct_pred_pos_curve(self, interval=False, delta_tau=0.001):
        """ Computes the relationship between precision
        and the percent of positively classified datapoints . """
        # compute thresholds based on the sorted probabilities
        orig_thresh = self.threshold
        sorted_labels, sorted_probs = self.sorted_values

        precisions = []
        pct_pred_pos = []
        taus = []
        tau = 0
        if not interval:
            for k in range(len(sorted_labels)):
                # compute new accuracy
                self.threshold = tau
                precisions.append(self.precision)
                pct_pred_pos.append(self.pct_pred_pos)
                taus.append(tau)
                
                # update threshold
                tau = sorted_probs[k]

        else:
            while tau < 1.0:
                # compute new accuracy
                self.threshold = tau
                precisions.append(self.precision)
                pct_pred_pos.append(self.pct_pred_pos)
                taus.append(tau)
                
                # update threshold
                tau += delta_tau

        # add last datapoint
        tau = 1.0
        self.threshold = tau
        precisions.append(self.precision)
        pct_pred_pos.append(self.pct_pred_pos)
        taus.append(tau)

        precisions.append(1.0)
        pct_pred_pos.append(0.0)
        taus.append(1.0 + 1e-12)

        self.threshold = orig_thresh
        return precisions, pct_pred_pos, taus




"""
Common geometric primitives.
Author: Jeff Mahler
"""
import numpy as np

class Box(object):
    """A 2D box or 3D rectangular prism.

    Attributes
    ----------
    dims : :obj:`numpy.ndarray` of float
        Maximal extent in x, y, and (optionally) z.

    width : float
        Maximal extent in x.

    height : float
        Maximal extent in y.

    area : float
        Area of projection onto xy plane.

    min_pt : :obj:`numpy.ndarray` of float
        The minimum x, y, and (optionally) z points.

    max_pt : :obj:`numpy.ndarray` of float
        The maximum x, y, and (optionally) z points.

    center : :obj:`numpy.ndarray` of float
        The center of the box in 2 or 3D coords.

    frame : :obj:`str`
        The frame in which this box is placed.
    """

    def __init__(self, min_pt, max_pt, frame='unspecified'):
        """Initialize a box.

        Parameters
        ----------
        min_pt : :obj:`numpy.ndarray` of float
            The minimum x, y, and (optionally) z points.

        max_pt : :obj:`numpy.ndarray` of float
            The maximum x, y, and (optionally) z points.

        frame : :obj:`str`
            The frame in which this box is placed.

        Raises
        ------
        ValueError
            If max_pt is not strictly larger than min_pt in all dims.
        """
        if np.any((max_pt - min_pt) < 0):
            raise ValueError('Min point must be smaller than max point')
        self._min_pt = min_pt
        self._max_pt = max_pt
        self._frame = frame

    @property
    def dims(self):
        """:obj:`numpy.ndarray` of float: Maximal extent in x, y, and (optionally) z
        """
        return self._max_pt - self._min_pt

    @property
    def width(self):
        """float: Maximal extent in x.
        """
        return int(np.round(self.dims[1]))

    @property
    def height(self):
        """float: Maximal extent in y.
        """
        return int(np.round(self.dims[0]))

    @property
    def area(self):
        """float: Area of projection onto xy plane.
        """
        return self.width * self.height

    @property
    def min_pt(self):
        """:obj:`numpy.ndarray` of float: The minimum x, y, and (optionally) z points.
        """
        return self._min_pt

    @property
    def max_pt(self):
        """:obj:`numpy.ndarray` of float: The maximum x, y, and (optionally) z points.
        """
        return self._max_pt

    @property
    def center(self):
        """:obj:`numpy.ndarray` of float: The center of the box in 2 or 3D coords.
        """
        return self.min_pt + self.dims / 2.0

    @property
    def ci(self):
        """float value of center i coordinate"""
        return self.center[0]

    @property
    def cj(self):
        """float value of center j coordinate"""
        return self.center[1]

    @property
    def frame(self):
        """:obj:`str`: The frame in which this box is placed.
        """
        return self._frame

class Contour(object):
    """ A set of pixels forming the boundary of an object of interest in an image.

    Attributes
    ----------
    boundary_pixels : :obj:`numpy.ndarray`
        Nx2 array of pixel coordinates on the boundary of a contour
    bounding_box : :obj:`Box`
        smallest box containing the contour
    area : float
        area of the contour
    num_pixels : int
        number of pixels along the boundary
    """
    def __init__(self, boundary_pixels, area=0.0, frame='unspecified'):
        self.boundary_pixels = boundary_pixels.squeeze()
        self.bounding_box = Box(np.min(self.boundary_pixels, axis=0),
                                np.max(self.boundary_pixels, axis=0),
                                frame)
        self.area = area
        
    @property
    def num_pixels(self):
        return self.boundary_pixels.shape[0]




"""
Common ops on lists / arrays of points
Author: Jeff Mahler
"""
from abc import ABCMeta, abstractmethod
import numbers

import numpy as np

from .primitives import Box

class BagOfPoints(object):
    """The abstract base class for collections of 3D point clouds.
    """
    __metaclass__ = ABCMeta

    def __init__(self, data, frame):
        """Initialize a BagOfPoints.

        Parameters
        ----------
        data : :obj:`numpy.ndarray` of float
            The data with which to initialize the collection. Usually is of
            shape (dim x #elements).
        frame : :obj:`str`
            The reference frame in which the collection of primitives
            resides.

        Raises
        ------
        ValueError
            If data is not a ndarray or frame is not a string.
        """
        if not isinstance(data, np.ndarray):
            raise ValueError('Must initialize bag of points with a numpy ndarray')
        if not isinstance(frame, str) and not isinstance(frame, unicode):
            raise ValueError('Must provide string name of frame of data')

        self._check_valid_data(data)
        self._data = self._preprocess_data(data)
        self._frame = frame

    @abstractmethod
    def _check_valid_data(self, data):
        """Checks that the data is valid for the appropriate class type.
        """
        pass

    def _preprocess_data(self, data):
        """Converts the data array to the preferred dim x #points structure.

        Parameters
        ----------
        data : :obj:`numpy.ndarray` of float
            The data to process.

        Returns
        -------
        :obj:`numpy.ndarray` of float
            The same data array, but reshapes lists to be dim x 1.
        """
        if len(data.shape) == 1: 
            data = data[:,np.newaxis]
        return data

    @property
    def shape(self):
        """:obj:`tuple` of int : The shape of the collection's data matrix.
        """
        return self._data.shape

    @property
    def frame(self):
        """:obj:`str` : The name of the reference frame in which the
        collection's objects reside.
        """
        return self._frame

    @property
    def data(self):
        """:obj:`numpy.ndarray` of float : The collection's data matrix.
        """
        return self._data.squeeze()

    @property
    def dim(self):
        """int : The number of entries in the data array along the first
        dimension. By convention, this is the dimension of the elements (usually
        3D).
        """
        return self._data.shape[0]

    @property
    def num_points(self):
        """int : The number of entries in the data array along the second
        dimenstion. By convention, this is the number of elements in the
        collection.
        """
        return self._data.shape[1]

    def copy(self):
        """Return a copy of the BagOfPoints object.

        Returns
        -------
        :obj:`BagOfPoints`
            An object of the same type as the original, with a copy
            of the data and the same frame.
        """
        return type(self)(self._data.copy(), self._frame)

    def save(self, filename):
        """Saves the collection to a file.

        Parameters
        ----------
        filename : :obj:`str`
            The file to save the collection to.

        Raises
        ------
        ValueError
            If the file extension is not .npy or .npz.
        """
        file_root, file_ext = os.path.splitext(filename)
        if file_ext == '.npy':
            np.save(filename, self._data)
        elif file_ext == '.npz':
            np.savez_compressed(filename, self._data)
        else:
            raise ValueError('Extension %s not supported for point saves.' %(file_ext))

    def load_data(filename):
        """Loads data from a file.

        Parameters
        ----------
        filename : :obj:`str`
            The file to load the collection from.

        Returns
        -------
        :obj:`numpy.ndarray` of float
            The data read from the file.

        Raises
        ------
        ValueError
            If the file extension is not .npy or .npz.
        """
        file_root, file_ext = os.path.splitext(filename)
        data = None
        if file_ext == '.npy':
            data = np.load(filename)
        elif file_ext == '.npz':
            data = np.load(filename)['arr_0']
        else:
            raise ValueError('Extension %s not supported for point reads' %(file_ext))
        return data

    def __getitem__(self, i):
        """Return a single element from the collection.

        Parameters
        ----------
        i : indexing-type (int or slice or list)
            The index of the desired element.

        Returns
        -------
        :obj:`Point` or :obj:`PointCloud`
            The returned element or group.
        """
        if isinstance(i, int):
            if i >= self.num_points:
                raise ValueError('Index %d is out of bounds' %(i))
            return Point(self._data[:,i], self._frame)
        if isinstance(i, list):
            i = np.array(i)
        if isinstance(i, np.ndarray):
            if np.max(i) >= self.num_points:
                raise ValueError('Index %d is out of bounds' %(np.max(i)))
            return PointCloud(self._data[:,i], self._frame)
        if isinstance(i, slice):
            return PointCloud(self._data[:,i], self._frame)
        raise ValueError('Type %s not supported for indexing' %(type(i)))

    def __str__(self):
        return str(self.data)

class BagOfVectors(BagOfPoints):
    """The base class for collections of 3D vectors.
    """
    pass

class Point(BagOfPoints):
    """A single 3D point.
    """

    def __init__(self, data, frame='unspecified'):
        """Initialize a Point.

        Parameters
        ----------
        data : :obj:`numpy.ndarray` of float
            An dim x 1 vector that represents the point's location.
        frame : :obj:`str`
            The reference frame in which the Point resides.
        """
        BagOfPoints.__init__(self, data, frame)

    def _check_valid_data(self, data):
        """Checks that the incoming data is a Nx1 ndarray.

        Parameters
        ----------
        data : :obj:`numpy.ndarray`
            The data to verify.

        Raises
        ------
        ValueError
            If the data is not of the correct shape.
        """
        if len(data.shape) == 2 and data.shape[1] != 1:
            raise ValueError('Can only initialize Point from a single Nx1 array') 

    @property
    def vector(self):
        """:obj:`numpy.ndarray` of float : The point's location relative
        to the origin.
        """
        return self._data.squeeze()

    @property
    def x(self):
        """float : The first element in the point's data array.
        """
        return self.vector[0]

    @property
    def y(self):
        """float : The second element in the point's data array.
        """
        return self.vector[1]

    @property
    def z(self):
        """float : The third element in the point's data array.
        """
        return self.vector[2]

    def __getitem__(self, dim):
        """float: The point value at the given dimension.
        """
        return self.vector[dim]
    
    def __add__(self, other_pt):
        """Add two Points together.

        Parameters
        ----------
        other_pt : :obj:`Point` or :obj:`numpy.ndarray` of float
            The other point to add to this one.

        Returns
        -------
        :obj:`Point`
            The result of adding the two points together.

        Raises
        ------
        ValueError
            If the shape and/or frames of the two points does not match.
        """
        # Handle point adds
        if isinstance(other_pt, Point) and other_pt.dim == self.dim:
            if self._frame != other_pt.frame:
                raise ValueError('Frames must be the same for addition')
            return Point(self.data + other_pt.data, frame=self._frame)
        # Handle numpy adds
        elif isinstance(other_pt, np.ndarray) and other_pt.shape == self.data.shape:
            return Point(self.data + other_pt, frame=self._frame)
        raise ValueError('Can only add to other Point objects or numpy ndarrays of the same dim')

    def __sub__(self, other_pt):
        """Subtract a point from this one.

        Parameters
        ----------
        other_pt : :obj:`Point` or :obj:`numpy.ndarray` of float
            The other point to subtract from this one.

        Returns
        -------
        :obj:`Point`
            The result of the subtraction.

        Raises
        ------
        ValueError
            If the shape and/or frames of the two points does not match.
        """
        return self + -1 * other_pt

    def __mul__(self, mult):
        """Multiply the point by a scalar.

        Parameters
        ----------
        mult : float
            The number by which to multiply the Point.

        Returns
        -------
        :obj:`Point3D`
            A 3D point created by the multiplication.

        Raises
        ------
        ValueError
            If mult is not a scalar value.
        """
        if isinstance(mult, numbers.Number):
            return Point(mult * self._data, self._frame)
        raise ValueError('Type %s not supported. Only scalar multiplication is supported' %(type(mult)))

    def __rmul__(self, mult):
        """Multiply the point by a scalar.

        Parameters
        ----------
        mult : float
            The number by which to multiply the Point.

        Returns
        -------
        :obj:`Point3D`
            A 3D point created by the multiplication.

        Raises
        ------
        ValueError
            If mult is not a scalar value.
        """
        return self.__mul__(mult)

    def __div__(self, div):
        """Divide the point by a scalar.

        Parameters
        ----------
        div : float
            The number by which to divide the Point.

        Returns
        -------
        :obj:`Point3D`
            A 3D point created by the division. 

        Raises
        ------
        ValueError
            If div is not a scalar value.
        """
        if not isinstance(div, numbers.Number):
            raise ValueError('Type %s not supported. Only scalar division is supported' %(type(div)))
        return self.__mul__(1.0 / div)

    @staticmethod
    def open(filename, frame='unspecified'):
        """Create a Point from data saved in a file.

        Parameters
        ----------
        filename : :obj:`str`
            The file to load data from.

        frame : :obj:`str`
            The frame to apply to the created point.

        Returns
        -------
        :obj:`Point`
            A point created from the data in the file.
        """
        data = BagOfPoints.load_data(filename)
        return Point(data, frame)

class Direction(BagOfVectors):
    """A single directional vector.
    """
    def __init__(self, data, frame):
        """Initialize a Direction.

        Parameters
        ----------
        data : :obj:`numpy.ndarray` of float
            An dim x 1 normalized vector.
        frame : :obj:`str`
            The reference frame in which the Direction resides.
        """
        BagOfPoints.__init__(self, data, frame)

    def _check_valid_data(self, data):
        """Checks that the incoming data is a Nx1 ndarray.

        Parameters
        ----------
        data : :obj:`numpy.ndarray`
            The data to verify.

        Raises
        ------
        ValueError
            If the data is not of the correct shape or if the vector is not
            normed.
        """
        if len(data.shape) == 2 and data.shape[1] != 1:
            raise ValueError('Can only initialize Direction from a single Nx1 array') 
        if np.abs(np.linalg.norm(data) - 1.0) > 1e-4:
            raise ValueError('Direction data must have norm=1.0')

    def orthogonal_basis(self):
        """Return an orthogonal basis to this direction.

        Note
        ----
            Only implemented in 3D.

        Returns
        -------
        :obj:`tuple` of :obj:`Direction`
            The pair of normalized Direction vectors that form a basis of
            this direction's orthogonal complement.

        Raises
        ------
        NotImplementedError
            If the vector is not 3D
        """
        if self.dim == 3:
            x_arr = np.array([-self.data[1], self.data[0], 0])
            if np.linalg.norm(x_arr) == 0:
                x_arr = np.array([self.data[2], 0, 0])
            x_arr = x_arr / np.linalg.norm(x_arr)
            y_arr = np.cross(self.data, x_arr)
            return Direction(x_arr, frame=self.frame), Direction(y_arr, frame=self.frame)
        raise NotImplementedError('Orthogonal basis only supported for 3 dimensions')

    @staticmethod
    def open(filename, frame='unspecified'):
        """Create a Direction from data saved in a file.

        Parameters
        ----------
        filename : :obj:`str`
            The file to load data from.

        frame : :obj:`str`
            The frame to apply to the created Direction.

        Returns
        -------
        :obj:`Direction`
            A Direction created from the data in the file.
        """
        data = BagOfPoints.load_data(filename)
        return Direction(data, frame)

class Plane3D(object):
    """A plane in three dimensions.
    """

    def __init__(self, n, x0):
        """Initialize a plane with a normal vector and a point.

        Parameters
        ----------
        n : :obj:`Direction`
            A 3D normal vector to the plane.

        x0 : :obj:`Point`
            A 3D point in the plane.

        Raises
        ------
        ValueError
            If the parameters are of the wrong type or are not of dimension 3.
        """
        if not isinstance(n, Direction) or n.dim != 3:
            raise ValueError('Plane normal must be a 3D direction')
        if not isinstance(x0, Point) or x0.dim != 3:
            raise ValueError('Plane offset must be a 3D point')
        self._n = n
        self._x0 = x0

    def split_points(self, point_cloud):
        """Split a point cloud into two along this plane.

        Parameters
        ----------
        point_cloud : :obj:`PointCloud`
            The PointCloud to divide in two.

        Returns
        -------
        :obj:`tuple` of :obj:`PointCloud`
            Two new PointCloud objects. The first contains points above the
            plane, and the second contains points below the plane.

        Raises
        ------
        ValueError
            If the input is not a PointCloud.
        """
        if not isinstance(point_cloud, PointCloud):
            raise ValueError('Can only split point clouds')
        # compute indices above and below
        above_plane = point_cloud._data - np.tile(self._x0.data, [1, point_cloud.num_points]).T.dot(self._n) > 0
        above_plane = point_cloud.z_coords > 0 & above_plane
        below_plane = point_cloud._data - np.tile(self._x0.data, [1, point_cloud.num_points]).T.dot(self._n) <= 0
        below_plane = point_cloud.z_coords > 0 & below_plane

        # split data
        above_data = point_cloud.data[:, above_plane]
        below_data = point_cloud.data[:, below_plane]
        return PointCloud(above_data, point_cloud.frame), PointCloud(below_data, point_cloud.frame)

class PointCloud(BagOfPoints):
    """A set of points.
    """

    def __init__(self, data, frame='unspecified'):
        """Initialize a PointCloud.

        Parameters
        ----------
        data : :obj:`numpy.ndarray` of float
            An dim x #points array that contains the points in the cloud.
        frame : :obj:`str`
            The reference frame in which the points reside.
        """
        BagOfPoints.__init__(self, data, frame)

    def _check_valid_data(self, data):
        """Checks that the incoming data is a 3 x #elements ndarray.

        Parameters
        ----------
        data : :obj:`numpy.ndarray`
            The data to verify.

        Raises
        ------
        ValueError
            If the data is not of the correct shape or type.
        """
        if data.dtype.type != np.float32 and data.dtype.type != np.float64:
            raise ValueError('Must initialize point clouds with a numpy float ndarray')
        if data.shape[0] != 3:
            raise ValueError('Illegal data array passed to point cloud. Must have 3 coordinates')
        if len(data.shape) > 2:
            raise ValueError('Illegal data array passed to point cloud. Must have 1 or 2 dimensions')

    @property
    def x_coords(self):
        """:obj:`numpy.ndarray` of float : An array containing all x coordinates
        in the cloud.
        """
        return self._data[0,:]

    @property
    def y_coords(self):
        """:obj:`numpy.ndarray` of float : An array containing all y coordinates
        in the cloud.
        """
        return self._data[1,:]

    @property
    def z_coords(self):
        """:obj:`numpy.ndarray` of float : An array containing all z coordinates
        in the cloud.
        """
        return self._data[2,:]

    def mean(self):
        """Returns the average point in the cloud.

        Returns
        -------
        :obj:`Point`
            The mean point in the PointCloud.
        """
        mean_point_data = np.mean(self._data, axis=1)
        return Point(mean_point_data, self._frame)

    def subsample(self, rate, random=False):
        """Returns a subsampled version of the PointCloud.

        Parameters
        ----------
        rate : int
            Only every rate-th element of the PointCloud is returned.

        Returns
        -------
        :obj:`PointCloud`
            A subsampled point cloud with N / rate total samples.

        Raises
        ------
        ValueError
            If rate is not a positive integer.
        """
        if type(rate) != int and rate < 1:
            raise ValueError('Can only subsample with strictly positive integer rate')
        indices = np.arange(self.num_points)
        if random:
            np.random.shuffle(indices)
        subsample_inds = indices[::rate]
        subsampled_data = self._data[:,subsample_inds]
        return PointCloud(subsampled_data, self._frame), subsample_inds

    def box_mask(self, box):
        """Return a PointCloud containing only points within the given Box.

        Parameters
        ----------
        box : :obj:`Box`
            A box whose boundaries are used to filter points.

        Returns
        -------
        :obj:`PointCloud`
            A filtered PointCloud whose points are all in the given box.
        :obj:`numpy.ndarray`
            Array of indices of the segmented points in the original cloud

        Raises
        ------
        ValueError
            If the input is not a box in the same frame as the PointCloud.
        """
        if not isinstance(box, Box):
            raise ValueError('Must provide Box object')
        if box.frame != self.frame:
            raise ValueError('Box must be in same frame as PointCloud')
        all_points = self.data.T
        cond1 = np.all(box.min_pt <= all_points, axis=1)
        cond2 = np.all(all_points <= box.max_pt, axis=1)
        valid_point_indices = np.where(np.logical_and(cond1, cond2))[0]
        valid_points = all_points[valid_point_indices]

        return PointCloud(valid_points.T, self.frame), valid_point_indices

    def best_fit_plane(self):
        """Fits a plane to the point cloud using least squares.

        Returns
        -------
        :obj:`tuple` of :obj:`numpy.ndarray` of float
            A normal vector to and point in the fitted plane.
        """
        X = np.c_[self.x_coords, self.y_coords, np.ones(self.num_points)]
        y = self.z_coords
        A = X.T.dot(X)
        b = X.T.dot(y)
        w = np.linalg.inv(A).dot(b)
        n = np.array([w[0], w[1], -1])
        n = n / np.linalg.norm(n)
        n = Direction(n, self._frame)
        x0 = self.mean()
        return n, x0

    def nonzero_indices(self):
        """ Returns the point indices corresponding to the zero points.
        
        Returns
        -------
        :obj:`numpy.ndarray`
            array of the nonzero indices
        """
        points_of_interest = np.where(self.z_coords != 0.0)[0]
        return points_of_interest
        
    def remove_zero_points(self):
        """Removes points with a zero in the z-axis.

        Note
        ----
        This returns nothing and updates the PointCloud in-place.
        """
        points_of_interest = np.where(self.z_coords != 0.0)[0]
        self._data = self.data[:, points_of_interest]

    def remove_infinite_points(self):
        """Removes infinite points.

        Note
        ----
        This returns nothing and updates the PointCloud in-place.
        """
        points_of_interest = np.where(np.all(np.isfinite(self.data), axis=0))[0]
        self._data = self.data[:, points_of_interest]
        
    def __add__(self, other_pc):
        """Add two PointClouds together element-wise.

        Parameters
        ----------
        other_pc : :obj:`PointCloud`
            The other PointCloud to add to this one.

        Returns
        -------
        :obj:`PointCloud`
            The result of adding the two PointClouds together.

        Raises
        ------
        ValueError
            If the shape and/or frames of the two PointClouds do not match.
        """
        if not isinstance(other_pc, PointCloud) or other_pc.num_points != self.num_points:
            raise ValueError('Can only add to other point clouds of same size')
        if self._frame != other_pc.frame:
            raise ValueError('Frames must be the same for addition')
        return PointCloud(self.data + other_pc.data, frame=self._frame)

    def __sub__(self, other_pc):
        """Subtract one PointCloud from another element-wise.

        Parameters
        ----------
        other_pc : :obj:`PointCloud`
            The other PointCloud to subtract from this one.

        Returns
        -------
        :obj:`PointCloud`
            The result of the subtraction.

        Raises
        ------
        ValueError
            If the shape and/or frames of the two PointClouds do not match.
        """
        return self + -1 * other_pc

    def __mul__(self, mult):
        """Multiply each point in the cloud by a scalar.

        Parameters
        ----------
        mult : float
            The number by which to multiply the PointCloud.

        Returns
        -------
        :obj:`PointCloud`
            A PointCloud created by the multiplication.

        Raises
        ------
        ValueError
            If mult is not a scalar value.
        """
        if isinstance(mult, numbers.Number):
            return PointCloud(mult * self._data, self._frame)
        raise ValueError('Type %s not supported. Only scalar multiplication is supported' %(type(mult)))

    def __rmul__(self, mult):
        """Multiply each point in the cloud by a scalar.

        Parameters
        ----------
        mult : float
            The number by which to multiply the PointCloud.

        Returns
        -------
        :obj:`PointCloud`
            A PointCloud created by the multiplication.

        Raises
        ------
        ValueError
            If mult is not a scalar value.
        """

        return self.__mul__(mult)

    def __div__(self, div):
        """Divide each point in the cloud by a scalar.

        Parameters
        ----------
        div : float
            The number by which to divide the PointCloud.

        Returns
        -------
        :obj:`PointCloud`
            A PointCloud created by the division.

        Raises
        ------
        ValueError
            If div is not a scalar value.
        """
        if not isinstance(div, numbers.Number):
            raise ValueError('Type %s not supported. Only scalar division is supported' %(type(div)))
        return self.__mul__(1.0 / div)

    @staticmethod
    def open(filename, frame='unspecified'):
        """Create a PointCloud from data saved in a file.

        Parameters
        ----------
        filename : :obj:`str`
            The file to load data from.

        frame : :obj:`str`
            The frame to apply to the created PointCloud.

        Returns
        -------
        :obj:`PointCloud`
            A PointCloud created from the data in the file.
        """
        data = BagOfPoints.load_data(filename)
        return PointCloud(data, frame)

class NormalCloud(BagOfVectors):
    """A set of normal vectors.
    """
    def __init__(self, data, frame='unspecified'):
        """Initialize a NormalCloud.

        Parameters
        ----------
        data : :obj:`numpy.ndarray` of float
            An dim x #points array that contains the normal vectors in the cloud.
            All of these vectors should be normalized.
        frame : :obj:`str`
            The reference frame in which the vectors reside.
        """
        BagOfPoints.__init__(self, data, frame)

    def _check_valid_data(self, data):
        """Checks that the incoming data is a 3 x #elements ndarray of normal
        vectors.

        Parameters
        ----------
        data : :obj:`numpy.ndarray`
            The data to verify.

        Raises
        ------
        ValueError
            If the data is not of the correct shape or type, or if the vectors
            therein are not normalized.
        """
        if data.dtype.type != np.float32 and data.dtype.type != np.float64:
            raise ValueError('Must initialize normals clouds with a numpy float ndarray')
        if data.shape[0] != 3:
            raise ValueError('Illegal data array passed to normal cloud. Must have 3 coordinates')
        if len(data.shape) > 2:
            raise ValueError('Illegal data array passed to normal cloud. Must have 1 or 2 dimensions')
        if np.any((np.abs(np.linalg.norm(data, axis=0) - 1) > 1e-4) & (np.linalg.norm(data, axis=0) != 0)):
            raise ValueError('Illegal data array passed to normal cloud. Must have norm=1.0 or norm=0.0')

    @property
    def x_coords(self):
        """:obj:`numpy.ndarray` of float : An array containing all x coordinates
        in the cloud.
        """
        return self._data[0,:]

    @property
    def y_coords(self):
        """:obj:`numpy.ndarray` of float : An array containing all y coordinates
        in the cloud.
        """
        return self._data[1,:]

    @property
    def z_coords(self):
        """:obj:`numpy.ndarray` of float : An array containing all z coordinates
        in the cloud.
        """
        return self._data[2,:]

    def subsample(self, rate):
        """Returns a subsampled version of the NormalCloud.

        Parameters
        ----------
        rate : int
            Only every rate-th element of the NormalCloud is returned.

        Returns
        -------
        :obj:`RateCloud`
            A subsampled point cloud with N / rate total samples.

        Raises
        ------
        ValueError
            If rate is not a positive integer.
        """
        if type(rate) != int and rate < 1:
            raise ValueError('Can only subsample with strictly positive integer rate')
        subsample_inds = np.arange(self.num_points)[::rate]
        subsampled_data = self._data[:,subsample_inds]
        return NormalCloud(subsampled_data, self._frame)

    def remove_zero_normals(self):
        """Removes normal vectors with a zero magnitude.

        Note
        ----
        This returns nothing and updates the NormalCloud in-place.
        """
        points_of_interest = np.where(np.linalg.norm(self._data, axis=0) != 0.0)[0]
        self._data = self._data[:, points_of_interest]

    def remove_nan_normals(self):
        """Removes normal vectors with nan magnitude.

        Note
        ----
        This returns nothing and updates the NormalCloud in-place.
        """
        points_of_interest = np.where(np.isfinite(np.linalg.norm(self._data, axis=0)))[0]
        self._data = self._data[:, points_of_interest]

    @staticmethod
    def open(filename, frame='unspecified'):
        """Create a NormalCloud from data saved in a file.

        Parameters
        ----------
        filename : :obj:`str`
            The file to load data from.

        frame : :obj:`str`
            The frame to apply to the created NormalCloud.

        Returns
        -------
        :obj:`NormalCloud`
            A NormalCloud created from the data in the file.
        """
        data = BagOfPoints.load_data(filename)
        return NormalCloud(data, frame)

class ImageCoords(BagOfPoints):
    """A set of 2D image coordinates.
    """

    def __init__(self, data, frame):
        """Initialize a set of image coodinates.

        Parameters
        ----------
        data : :obj:`numpy.ndarray` of int
            An 2 x #coords array that contains the image coordinates.
        frame : :obj:`str`
            The reference frame in which the points reside.
        """
        BagOfPoints.__init__(self, data, frame)

    def _check_valid_data(self, data):
        """Checks that the incoming data is a 2 x #elements ndarray of ints.

        Parameters
        ----------
        data : :obj:`numpy.ndarray`
            The data to verify.

        Raises
        ------
        ValueError
            If the data is not of the correct shape or type.
        """
        if data.dtype.type != np.int8 and data.dtype.type != np.int16 \
                and data.dtype.type != np.int32 and data.dtype.type != np.int64 \
                and data.dtype.type != np.uint8 and data.dtype.type != np.uint16 \
                and data.dtype.type != np.uint32 and data.dtype.type != np.uint64:
            raise ValueError('Must initialize image coords with a numpy int ndarray')
        if data.shape[0] != 2:
            raise ValueError('Illegal data array passed to image coords. Must have 2 coordinates')
        if len(data.shape) > 2:
            raise ValueError('Illegal data array passed to point cloud. Must have 1 or 2 dimensions')

    @property
    def i_coords(self):
        """:obj:`numpy.ndarray` of float : The set of i-coordinates
        (those in the second row of the data matrix).
        """
        return self._data[1,:]

    @property
    def j_coords(self):
        """:obj:`numpy.ndarray` of float : The set of j-coordinates
        (those in the first row of the data matrix).
        """
        return self._data[0,:]

    @staticmethod
    def open(filename, frame='unspecified'):
        """Create an ImageCoords from data saved in a file.

        Parameters
        ----------
        filename : :obj:`str`
            The file to load data from.

        frame : :obj:`str`
            The frame to apply to the created ImageCoords.

        Returns
        -------
        :obj:`ImageCoords`
            An ImageCoords created from the data in the file.
        """
        data = BagOfPoints.load_data(filename)
        return ImageCoords(data, frame)

class RgbCloud(BagOfPoints):
    """A set of colors.
    """

    def __init__(self, data, frame):
        """Initialize an RgbCloud.

        Parameters
        ----------
        data : :obj:`numpy.ndarray` of uint8
            An 3 x #elements array that contains the colors in the cloud.
            Elements each have a red, green, and blue coordinate.
        frame : :obj:`str`
            The reference frame in which the points reside.
        """
        BagOfPoints.__init__(self, data, frame)

    def _check_valid_data(self, data):
        """Checks that the incoming data is a 3 x #elements ndarray.

        Parameters
        ----------
        data : :obj:`numpy.ndarray`
            The data to verify.

        Raises
        ------
        ValueError
            If the data is not of the correct shape or type.
        """
        if data.dtype.type != np.uint8:
            raise ValueError('Must initialize rgb clouds with a numpy uint ndarray')
        if data.shape[0] != 3:
            raise ValueError('Illegal data array passed to rgb cloud. Must have 3 coordinates')
        if len(data.shape) > 2:
            raise ValueError('Illegal data array passed to rgb  cloud. Must have 1 or 2 dimensions')

    @property
    def red(self):
        """:obj:`numpy.ndarray` of uint8 : An array containing all red values
        in the cloud.
        """
        return self._data[0,:]

    @property
    def green(self):
        """:obj:`numpy.ndarray` of uint8 : An array containing all green values
        in the cloud.
        """
        return self._data[1,:]

    @property
    def blue(self):
        """:obj:`numpy.ndarray` of uint8 : An array containing all blue values
        in the cloud.
        """
        return self._data[2,:]

    @staticmethod
    def open(filename, frame='unspecified'):
        """Create a RgbCloud from data saved in a file.

        Parameters
        ----------
        filename : :obj:`str`
            The file to load data from.

        frame : :obj:`str`
            The frame to apply to the created RgbCloud.

        Returns
        -------
        :obj:`RgbCloud`
            A RgdCloud created from the data in the file.
        """
        data = BagOfPoints.load_data(filename)
        return RgbCloud(data, frame)

class RgbPointCloud(object):
    """A combined set of 3D points and RGB colors.
    """

    def __init__(point_data, rgb_data, frame):
        """Initialize a PointCloud + RgbCloud combination.

        Parameters
        ----------
        point_data : :obj:`numpy.ndarray` of float
            An dim x #elements array that contains the points in the cloud.
        rgb_data : :obj:`numpy.ndarray` of uint8
            A 3 x #elements array that contains the colors of the cloud.
        frame : :obj:`str`
            The reference frame in which the points reside.
        """
        self.point_cloud = PointCloud(point_data, frame)
        self.rgb_cloud = RgbCloud(rgb_data, frame)

    def __getitem__(self, i):
        """Returns the ith point and color.

        Returns
        -------
        :obj:`tuple` of :obj:`numpy.ndarray` of float and uint8
            The ith point and the ith color.
        """
        return self.point_cloud[i], self.rgb_cloud[i]

class PointNormalCloud(object):
    """A combined set of 3D points and normal vectors.
    """

    def __init__(self, point_data, normal_data, frame):
        """Initialize a PointCloud + NormalCloud combination.

        Parameters
        ----------
        point_data : :obj:`numpy.ndarray` of float
            An dim x #elements array that contains the points in the cloud.
        normal_data : :obj:`numpy.ndarray` of uint8
            A dim x #elements array that contains the normals of the cloud.
        frame : :obj:`str`
            The reference frame in which the points reside.

        Raises
        ------
        ValueError
            If the two datasets don't have the same number of elements.
        """
        self.point_cloud = PointCloud(point_data, frame)
        self.normal_cloud = NormalCloud(normal_data, frame)
        if self.point_cloud.num_points != self.normal_cloud.num_points:
            raise ValueError('PointCloud and NormalCloud must have the same number of points')

    @property
    def points(self):
        """:obj:`PointCloud` : The PointCloud in this set.
        """
        return self.point_cloud

    @property
    def normals(self):
        """:obj:`NormalCloud` : The NormalCloud in this set.
        """
        return self.normal_cloud

    @property
    def num_points(self):
        """int : The number of elements in the clouds.
        """
        return self.point_cloud.num_points

    @property
    def frame(self):
        """:obj:`str` : The frame in which these clouds exist.
        """
        return self.point_cloud.frame

    def __getitem__(self, i):
        """Returns the ith point and normals.

        Returns
        -------
        :obj:`tuple` of :obj:`numpy.ndarray` of float 
            The ith point and the ith normal.
        """
        return self.point_cloud[i], self.normal_cloud[i]

    def remove_zero_points(self):
        """Remove all elements where the norms and points are zero.

        Note
        ----
        This returns nothing and updates the NormalCloud in-place.
        """
        points_of_interest = np.where((np.linalg.norm(self.point_cloud.data, axis=0) != 0.0)  &
                                      (np.linalg.norm(self.normal_cloud.data, axis=0) != 0.0) &
                                      (np.isfinite(self.normal_cloud.data[0,:])))[0]
        self.point_cloud._data = self.point_cloud.data[:, points_of_interest]
        self.normal_cloud._data = self.normal_cloud.data[:, points_of_interest]




'''
Class to handle experiment logging.
Authors: Jeff, Jacky
'''
from abc import ABCMeta, abstractmethod
import os
import csv
import shutil
import subprocess
from datetime import datetime
from time import time
import logging

import numpy as np

from .csv_model import CSVModel
from .yaml_config import YamlConfig
from .utils import gen_experiment_id

class ExperimentLogger:
    """Abstract class for experiment logging.

    Experiments are logged to CSV files, which are encapsulated with the
    :obj:`CSVModel` class.
    """
    __metaclass__ = ABCMeta

    _MASTER_RECORD_FILENAME = 'experiment_record.csv'

    def __init__(self, experiment_root_path, experiment_tag='experiment', log_to_file=True, sub_experiment_dirs=True):
        """Initialize an ExperimentLogger.

        Parameters
        ----------
        experiment_root_path : :obj:`str`
            The root directory in which to save experiment files.
        experiment_tag : :obj:`str`
            The tag to use when prefixing new experiments
        log_to_file : bool, optional
            Default: True
            If True will log all logging statements to a log file
        sub_experiment_dirs : bool, optional
            Defautl: True
            If True will make sub directories corresponding to generated experiment name
        """
        self.experiment_root_path = experiment_root_path

        # open the master record
        self.master_record_filepath = os.path.join(self.experiment_root_path, ExperimentLogger._MASTER_RECORD_FILENAME)
        self.master_record = CSVModel.get_or_create(self.master_record_filepath, self.experiment_meta_headers)

        # add new experiment to the master record
        self.id = ExperimentLogger.gen_experiment_ref(experiment_tag)
        self._master_record_uid = self.master_record.insert(self.experiment_meta_data)

        # make experiment output dir
        if sub_experiment_dirs:
            self.experiment_path = os.path.join(self.experiment_root_path, self.id)
            if not os.path.exists(self.experiment_path):
                os.makedirs(self.experiment_path)
        else:
            self.experiment_path = self.experiment_root_path

        if log_to_file:
            # redirect logging statements to a file
            if not sub_experiment_dirs:
                self.log_path = os.path.join(self.experiment_root_path, 'logs')
            else:
                self.log_path = self.experiment_path
            if not os.path.exists(self.log_path):
                os.makedirs(self.log_path)
            experiment_log = os.path.join(self.log_path, '%s.log' %(self.id))
            formatter = logging.Formatter('%(asctime)s %(levelname)s: %(message)s')
            hdlr = logging.FileHandler(experiment_log)
            hdlr.setFormatter(formatter)
            logging.getLogger().addHandler(hdlr)

        # internal dir struct
        self._dirs = {}

    @staticmethod
    def gen_experiment_ref(experiment_tag, n=10):
        """ Generate a random string for naming.

        Parameters
        ----------
        experiment_tag : :obj:`str`
            tag to prefix name with
        n : int
            number of random chars to use

        Returns
        -------
        :obj:`str`
            string experiment ref
        """
        experiment_id = gen_experiment_id(n=n)
        return '{0}_{1}'.format(experiment_tag, experiment_id)

    def update_master_record(self, data):
        """Update a row of the experimental master record CSV with the given data.

        Parameters
        ----------
        uid : int
            The UID of the row to update.

        data : :obj:`dict`
            A dictionary mapping keys (header strings) to values, which
            represents the new row.
        """
        self.master_record.update_by_uid(self._master_record_uid, data)

    @abstractmethod
    def experiment_meta_headers(self):
        """Returns list of two-tuples of header names and types of meta information for the experiments

        Returns
        -------
        :obj:`tuple`
            The metadata for this experiment.
        """
        pass

    @abstractmethod
    def experiment_meta_data(self):
        """Returns the dict of header names and value of meta information for the experiments

        Returns
        -------
        :obj:`dict`
            The metadata for this experiment.
        """
        pass

    @property
    def dirs(self):
        return self._dirs.copy()

    def construct_internal_dirs(self, dirs, realize=False):
        cur_dir = self._dirs
        for dir in dirs:
            if dir not in cur_dir:
                cur_dir[dir] = {}
            cur_dir = cur_dir[dir]
        if realize:
            self._realize_dirs(dirs)

    def construct_internal_dirs_group(self, group_dirs):
        for dirs in group_dirs:
            self.construct_internal_dirs(dirs)

    def has_internal_dirs(self, dirs):
        cur_dir = self.dirs
        for dir in dirs:
            if dir not in cur_dir:
                return False
            cur_dir = cur_dir[dir]
        return True

    def dirs_to_path(self, dirs):
        rel_path = '/'.join(dirs)
        abs_path = os.path.join(self.experiment_path, rel_path)
        return abs_path

    def _realize_dirs(self, dirs):
        if not self.has_internal_dirs(dirs):
            raise Exception("Directory has not been constructed internally! {0}".format(dirs))
        abs_path = self.dirs_to_path(dirs)
        if not os.path.exists(abs_path):
            os.makedirs(abs_path)
        return abs_path

    def remove_dirs(self, dirs):
        if not self.has_internal_dirs(dirs):
            raise Exception("Directory has not been construted internally! {0}".format(dirs))

        path = self.dirs_to_path(dirs)
        if os.path.exists(path):
            subprocess.call(['trash', '-r', path])

        # remove the deepest node
        cur_dir = self.dirs
        for dir in dirs[:-1]:
            cur_dir = cur_dir[dir]
        cur_dir.pop(dirs[-1])

        for i in range(len(dirs) - 1):
            cur_dir = self._dirs
            depth = len(dirs) - i - 2
            for j in range(depth):
                cur_dir = cur_dir[dirs[j]]

            dir_to_remove = dirs[depth]
            if not cur_dir[dir_to_remove]:
                cur_dir.pop(dir_to_remove)
            else:
                break

    def copy_to_dir(self, src_file_path, target_dirs):
        abs_path = self._realize_dirs(target_dirs)
        basename = os.path.basename(src_file_path)
        target_file_path = os.path.join(abs_path, basename)

        logging.debug("Copying {0} to {1}".format(src_file_path, target_file_path))
        shutil.copyfile(src_file_path, target_file_path)

    def copy_dirs(self, src_dirs_path, target_dirs):
        if not self.has_internal_dirs(target_dirs):
            raise Exception("Directory has not been constructed internally! {0}".format(target_dirs))

        target_dirs_path = self.dirs_to_path(target_dirs)
        if os.path.exists(target_dirs_path):
            if len(os.listdir(target_dirs_path)) > 0:
                raise Exception("Target path for copying directories is not empty! Got: {0}".format(target_dirs_path))
            else:
                os.rmdir(target_dirs_path)
        shutil.copytree(src_dirs_path, target_dirs_path)

    @staticmethod
    def pretty_str_time(dt):
        return "{0}_{1}_{2}_{3}:{4}".format(dt.year, dt.month, dt.day, dt.hour, dt.minute)



"""
Class to handle dual quaternions and their interpolations
Implementation details inspired by Ben Kenwright's "A Beginners Guide to Dual-Quaternions"
http://cs.gmu.edu/~jmlien/teaching/cs451/uploads/Main/dual-quaternion.pdf
Author: Jacky Liang
"""
from numbers import Number
import numpy as np

from .transformations import quaternion_multiply, quaternion_conjugate

class DualQuaternion(object):
    """Class for handling dual quaternions and their interpolations.

    Attributes
    ----------
    qr : :obj:`numpy.ndarray` of float
        A 4-entry quaternion in wxyz format.

    qd : :obj:`numpy.ndarray` of float
        A 4-entry quaternion in wxyz format.

    conjugate : :obj:`DualQuaternion`
        The conjugate of this DualQuaternion.

    norm : :obj:`tuple` of :obj:`numpy.ndarray`
        The normalized vectors for qr and qd, respectively.

    normalized : :obj:`DualQuaternion`
        This quaternion with qr normalized.
    """

    def __init__(self, qr=[1,0,0,0], qd=[0,0,0,0], enforce_unit_norm=True):
        """Initialize a dual quaternion.

        Parameters
        ----------
        qr : :obj:`numpy.ndarray` of float
            A 4-entry quaternion in wxyz format.

        qd : :obj:`numpy.ndarray` of float
            A 4-entry quaternion in wxyz format.

        enforce_unit_norm : bool
            If true, raises a ValueError when the quaternion is not normalized.

        Raises
        ------
        ValueError
            If enforce_unit_norm is True and the norm of qr is not 1.
        """
        self.qr = qr
        self.qd = qd

        if enforce_unit_norm:
            norm = self.norm
            if not np.allclose(norm[0], [1]):
                raise ValueError("Dual quaternion does not have norm 1! Got {0}".format(norm[0]))

    @property
    def qr(self):
        """:obj:`numpy.ndarray` of float: A 4-entry quaternion in wxyz format.
        """
        qr_wxyz = np.roll(self._qr, 1)
        return qr_wxyz

    @qr.setter
    def qr(self, qr_wxyz):
        qr_wxyz = np.array([n for n in qr_wxyz])
        qr_xyzw = np.roll(qr_wxyz, -1)
        self._qr = qr_xyzw

    @property
    def qd(self):
        """:obj:`numpy.ndarray` of float: A 4-entry quaternion in wxyz format.
        """
        qd_wxyz = np.roll(self._qd, 1)
        return qd_wxyz

    @qd.setter
    def qd(self, qd_wxyz):
        qd_wxyz = np.array([n for n in qd_wxyz])
        if qd_wxyz[0] != 0:
            raise ValueError('Invalid dual quaternion! First value of Qd must be 0. Got {0}'.format(qd))
        qd_xyzw = np.roll(qd_wxyz, -1)
        self._qd = qd_xyzw

    @property
    def conjugate(self):
        """:obj:`DualQuaternion`: The conjugate of this quaternion.
        """
        qr_c_xyzw = quaternion_conjugate(self._qr)
        qd_c_xyzw = quaternion_conjugate(self._qd)

        qr_c_wxyz = np.roll(qr_c_xyzw, 1)
        qd_c_wxyz = np.roll(qd_c_xyzw, 1)
        return DualQuaternion(qr_c_wxyz, qd_c_wxyz)

    @property
    def norm(self):
        """:obj:`tuple` of :obj:`numpy.ndarray`: The normalized vectors for qr and qd, respectively.
        """
        qr_c = quaternion_conjugate(self._qr)
        qd_c = quaternion_conjugate(self._qd)

        qr_norm = np.linalg.norm(quaternion_multiply(self._qr, qr_c))
        qd_norm = np.linalg.norm(quaternion_multiply(self._qr, qd_c) + quaternion_multiply(self._qd, qr_c))

        return (qr_norm, qd_norm)

    @property
    def normalized(self):
        """:obj:`DualQuaternion`: This quaternion with qr normalized.
        """
        qr = self.qr /1./ np.linalg.norm(self.qr)
        return DualQuaternion(qr, self.qd, True)

    def copy(self):
        """Return a copy of this quaternion.

        Returns
        -------
        :obj:`DualQuaternion`
            The copied DualQuaternion.
        """
        return DualQuaternion(self.qr.copy(), self.qd.copy())

    @staticmethod
    def interpolate(dq0, dq1, t):
        """Return the interpolation of two DualQuaternions.

        This uses the Dual Quaternion Linear Blending Method as described by Matthew Smith's
        'Applications of Dual Quaternions in Three Dimensional Transformation and Interpolation'
        https://www.cosc.canterbury.ac.nz/research/reports/HonsReps/2013/hons_1305.pdf

        Parameters
        ----------
        dq0 : :obj:`DualQuaternion`
            The first DualQuaternion.

        dq1 : :obj:`DualQuaternion`
            The second DualQuaternion.

        t : float
            The interpolation step in [0,1]. When t=0, this returns dq0, and
            when t=1, this returns dq1.

        Returns
        -------
        :obj:`DualQuaternion`
            The interpolated DualQuaternion.

        Raises
        ------
        ValueError
            If t isn't in [0,1].
        """
        if not 0 <= t <= 1:
            raise ValueError("Interpolation step must be between 0 and 1! Got {0}".format(t))

        dqt = dq0 * (1-t) + dq1 * t
        return dqt.normalized

    def __mul__(self, val):
        """Multiplies the dual quaternion by another dual quaternion or a
        scalar.

        Parameters
        ----------
        val : :obj:`DualQuaternion` or number
            The value by which to multiply this dual quaternion.

        Returns
        -------
        :obj:`DualQuaternion`
            A new DualQuaternion that results from the multiplication.

        Raises
        ------
        ValueError
            If val is not a DualQuaternion or Number.
        """
        if isinstance(val, DualQuaternion):
            new_qr_xyzw = quaternion_multiply(self._qr, val._qr)
            new_qd_xyzw = quaternion_multiply(self._qr, val._qd) + quaternion_multiply(self._qd, val._qr)

            new_qr_wxyz = np.roll(new_qr_xyzw, 1)
            new_qd_wxyz = np.roll(new_qd_xyzw, 1)

            return DualQuaternion(new_qr_wxyz, new_qd_wxyz)
        elif isinstance(val, Number):
            new_qr_wxyz = val * self.qr
            new_qd_wxyz = val * self.qd

            return DualQuaternion(new_qr_wxyz, new_qd_wxyz, False)

        raise ValueError('Cannot multiply dual quaternion with object of type {0}'.format(type(val)))

    def __add__(self, val):
        """Adds the dual quaternion to another dual quaternion.

        Parameters
        ----------
        val : :obj:`DualQuaternion`
            The DualQuaternion to add to this one.

        Returns
        -------
        :obj:`DualQuaternion`
            A new DualQuaternion that results from the addition..

        Raises
        ------
        ValueError
            If val is not a DualQuaternion.
        """
        if not isinstance(val, DualQuaternion):
            raise ValueError('Cannot add dual quaternion with object of type {0}'.format(type(val)))

        new_qr_wxyz = self.qr + val.qr
        new_qd_wxyz = self.qd + val.qd
        new_qr_wxyz = new_qr_wxyz / np.linalg.norm(new_qr_wxyz)

        return DualQuaternion(new_qr_wxyz, new_qd_wxyz, False)

    def __str__(self):
        return '{0}+{1}e'.format(self.qr, self.qd)

    def __repr__(self):
        return 'DualQuaternion({0},{1})'.format(repr(self.qr), repr(self.qd))




"""
Utility class for logging.

Author: Vishal Satish
"""
import logging
import sys

import colorlog

ROOT_LOG_LEVEL = logging.INFO
ROOT_LOG_STREAM = sys.stdout

def configure_root():
    """Configure the root logger."""
    root_logger = logging.getLogger()

    # clear any existing handles to streams because we don't want duplicate logs
    # NOTE: we assume that any stream handles we find are to ROOT_LOG_STREAM, which is usually the case(because it is stdout). This is fine because we will be re-creating that handle. Otherwise we might be deleting a handle that won't be re-created, which could result in dropped logs.
    for hdlr in root_logger.handlers:
        if isinstance(hdlr, logging.StreamHandler):
            root_logger.removeHandler(hdlr)

    # configure the root logger
    root_logger.setLevel(ROOT_LOG_LEVEL)
    hdlr = logging.StreamHandler(ROOT_LOG_STREAM) 
    formatter = colorlog.ColoredFormatter(
                        '%(purple)s%(name)-10s %(log_color)s%(levelname)-8s%(reset)s %(white)s%(message)s',
                        reset=True,
                        log_colors={
                            'DEBUG': 'cyan',
                            'INFO': 'green',
                            'WARNING': 'yellow',
                            'ERROR': 'red',
                            'CRITICAL': 'red,bg_white',
                        } 
                                         )   
    hdlr.setFormatter(formatter)
    root_logger.addHandler(hdlr)

def add_root_log_file(log_file):
    """
    Add a log file to the root logger.

    Parameters
    ----------
    log_file :obj:`str`
        The path to the log file.
    """
    root_logger = logging.getLogger()

    # add a file handle to the root logger
    hdlr = logging.FileHandler(log_file)
    formatter = logging.Formatter('%(asctime)s %(name)-10s %(levelname)-8s %(message)s', datefmt='%m-%d %H:%M:%S')
    hdlr.setFormatter(formatter)
    root_logger.addHandler(hdlr)
    root_logger.info('Root logger now logging to {}'.format(log_file))

class Logger(object):
    ROOT_CONFIGURED = False

    @staticmethod
    def reconfigure_root():
        """Reconfigure the root logger."""
        configure_root()

    @staticmethod
    def get_logger(name, log_level=logging.INFO, log_file=None, global_log_file=False, silence=False):
        """
        Build a logger. All logs will be propagated up to the root logger if not silenced. If log_file is provided, logs will be written out to that file. If global_log_file is true, log_file will be handed the root logger, otherwise it will only be used by this particular logger.
    
        Parameters
        ----------
        name :obj:`str`
            The name of the logger to be built.
        log_level : `int`
            The log level. See the python logging module documentation for possible enum values.
        log_file :obj:`str`
            The path to the log file to log to.
        global_log_file :obj:`bool`
            Whether or not to use the given log_file for this particular logger or for the root logger.
        silence :obj:`bool`
            Whether or not to silence this logger. If it is silenced, the only way to get output from this logger is through a non-global log file.

        Returns
        -------
        :obj:`logging.Logger`
            A custom logger.
        """
        no_op = False
        # some checks for silencing/no-op logging
        if silence and global_log_file:
            raise ValueError("You can't silence a logger and log to a global log file!")
        if silence and log_file is None:
            logging.warning('You are creating a no-op logger!')
            no_op = True

        # configure the root logger if it hasn't been already
        if not Logger.ROOT_CONFIGURED:
            configure_root()
        Logger.ROOT_CONFIGURED = True
        
        # build a logger
        logger = logging.getLogger(name)
        logger.setLevel(log_level)

        # silence the logger by preventing it from propagating upwards to the root
        logger.propagate = not silence

        # configure the log file stream 
        if log_file is not None:
            # if the log file is global, add it to the root logger
            if global_log_file:
                add_root_log_file(log_file)
            # otherwise add it to this particular logger
            else:
                hdlr = logging.FileHandler(log_file)
                formatter = logging.Formatter('%(asctime)s %(name)-10s %(levelname)-8s %(message)s', datefmt='%m-%d %H:%M:%S')
                hdlr.setFormatter(formatter)
                logger.addHandler(hdlr)

        # add a no-op handler to suppress warnings about there being no handlers
        if no_op:
            logger.addHandler(logging.NullHandler())
        return logger

    @staticmethod
    def add_log_file(logger, log_file, global_log_file=False):
        """
        Add a log file to this logger. If global_log_file is true, log_file will be handed the root logger, otherwise it will only be used by this particular logger.

        Parameters
        ----------
        logger :obj:`logging.Logger`
            The logger.
        log_file :obj:`str`
            The path to the log file to log to.
        global_log_file :obj:`bool`
            Whether or not to use the given log_file for this particular logger or for the root logger.
        """

        if global_log_file:
            add_root_log_file(log_file)
        else:
            hdlr = logging.FileHandler(log_file)
            formatter = logging.Formatter('%(asctime)s %(name)-10s %(levelname)-8s %(message)s', datefmt='%m-%d %H:%M:%S')
            hdlr.setFormatter(formatter)
            logger.addHandler(hdlr)



"""
Custom exception types.
Author: Jeff Mahler
"""
class TerminateException(Exception):
    """ Signal to terminate.
    """
    def __init__(self, *args, **kwargs):
         Exception.__init__(self, *args, **kwargs)





"""
Module to serialize and deserialize JSON with numpy arrays. Adapted from
http://stackoverflow.com/a/24375113/723090 so that arrays are human-readable.

Author: Brian Hou
"""

import json as _json
import numpy as np

class NumpyEncoder(_json.JSONEncoder):
    """A numpy array to json encoder.
    """

    def default(self, obj):
        """Converts an ndarray into a dictionary for efficient serialization.

        The dict has three keys:
        - dtype : The datatype of the array as a string.
        - shape : The shape of the array as a tuple.
        - __ndarray__ : The data of the array as a list.

        Parameters
        ----------
        obj : :obj:`numpy.ndarray`
            The ndarray to encode.

        Returns
        -------
        :obj:`dict`
            The dictionary serialization of obj.
        
        Raises
        ------
        TypeError
            If obj isn't an ndarray.
        """
        if isinstance(obj, np.ndarray):
            return dict(__ndarray__=obj.tolist(),
                        dtype=str(obj.dtype),
                        shape=obj.shape)
        # Let the base class default method raise the TypeError
        return _json.JSONEncoder(self, obj)

def json_numpy_obj_hook(dct):
    """Decodes a previously encoded numpy ndarray with proper shape and dtype.

    Parameters
    ----------
    dct : :obj:`dict`
        The encoded dictionary.

    Returns
    -------
    :obj:`numpy.ndarray`
        The ndarray that `dct` was encoding.
    """
    if isinstance(dct, dict) and '__ndarray__' in dct:
        data = np.asarray(dct['__ndarray__'], dtype=dct['dtype'])
        return data.reshape(dct['shape'])
    return dct

def dump(*args, **kwargs):
    """Dump a numpy.ndarray to file stream.

    This works exactly like the usual `json.dump()` function,
    but it uses our custom serializer.
    """
    kwargs.update(dict(cls=NumpyEncoder,
                       sort_keys=True,
                       indent=4,
                       separators=(',', ': ')))
    return _json.dump(*args, **kwargs)

def load(*args, **kwargs):
    """Load an numpy.ndarray from a file stream.

    This works exactly like the usual `json.load()` function,
    but it uses our custom deserializer.
    """
    kwargs.update(dict(object_hook=json_numpy_obj_hook))
    return _json.load(*args, **kwargs)



"""
Class to sync and rate limit multiple DataStreamRecorders
Author: Jacky Liang
"""
from multiprocess import Process, Queue
from time import time
try:
    from Queue import Empty
except:
    from queue import Empty
import logging, sys
from setproctitle import setproctitle

class _DataStreamSyncer(Process):

    def __init__(self, frequency, ok_qs, cmds_q, tokens_q):
        Process.__init__(self)
        self._cmds_q = cmds_q
        self._tokens_q = tokens_q

        self._ok_qs = ok_qs
        self._tokens = {key : False for key in self._ok_qs.keys()}

        self._T = 1./frequency if frequency > 0 else 0
        self._ok_start_time = None
        self._pause = False

    def run(self):
        self._session_start_time = time()
        setproctitle("python._DataStreamSyncer")
        try:
            while True:
                if not self._cmds_q.empty():
                    cmd = self._cmds_q.get()
                    if cmd[0] == "reset_time":
                        self._session_start_time = time()
                    elif cmd[0] == 'pause':
                        self._pause = True
                        self._pause_time = time()
                        self._take_oks()
                    elif cmd[0] == 'resume':
                        self._pause = False
                        if cmd[1]:
                            self._session_start_time = time()
                        else:
                            self._session_start_time += time() - self._pause_time
                    elif cmd[0] == "stop":
                        break
                if not self._tokens_q.empty():
                    motion, name = self._tokens_q.get()
                    if motion == "take":
                        self._tokens[name] = False
                    elif motion == "return":
                        self._tokens[name] = True
                self._try_ok()
        except KeyboardInterrupt:
            logging.debug("Exiting DataStreamSyncer")
            sys.exit(0)

    def _send_oks(self):
        self._ok_start_time = time()
        t = self._ok_start_time - self._session_start_time
        for ok_q in self._ok_qs.values():
            ok_q.put(t)

    def _take_oks(self):
        for ok_q in self._ok_qs.values():
            while ok_q.qsize() > 0:
                try:
                    ok_q.get_nowait()
                except Empty:
                    pass

    def _try_ok(self):
        if self._pause:
            return
        if False in self._tokens.values():
            if self._ok_start_time is not None and time() - self._ok_start_time > self._T:
                timeout_names = []
                for name, returned in self._tokens.items():
                    if not returned:
                        timeout_names.append(name)
                logging.warn("Potential timeout! {} not yet returned within desired period!".format(timeout_names))
            return
        if self._T <= 0 or self._ok_start_time is None or \
            time() - self._ok_start_time > self._T:
            self._send_oks()

class DataStreamSyncer:

    def __init__(self, data_stream_recorders, frequency=0):
        """
        Instantiates a new DataStreamSyncer

        Parameters
        ----------
            data_stream_recorders : list of DataStreamRecorders to sync
            frequency : float, optional
                    Frequency in hz used for ratelimiting. If set to 0 or less, will not rate limit.
                    Defaults to 0
        """
        self._cmds_q = Queue()
        self._tokens_q = Queue()

        self._data_stream_recorders = data_stream_recorders
        ok_qs = {}
        for data_stream_recorder in self._data_stream_recorders:
            ok_q = Queue()
            name = data_stream_recorder.name
            if name in ok_qs:
                raise ValueError("Data Stream Recorders must have unique names! {} is a duplicate!".format(name))
            ok_qs[name] = ok_q
            data_stream_recorder._set_qs(ok_q, self._tokens_q)

        self._syncer = _DataStreamSyncer(frequency, ok_qs, self._cmds_q, self._tokens_q)
        self._syncer.start()

    def start(self):
        """ Starts syncer operations """
        for recorder in self._data_stream_recorders:
            recorder._start_recording()

    def stop(self):
        """ Stops syncer operations. Destroys syncer process. """
        self._cmds_q.put(("stop",))
        for recorder in self._data_stream_recorders:
            recorder._stop()
        try:
            self._syncer.terminate()
        except Exception:
            pass

    def pause(self):
        self._cmds_q.put(("pause",))
        for recorder in self._data_stream_recorders:
            recorder._pause()

    def resume(self, reset_time=False):
        self._cmds_q.put(("resume",reset_time))
        for recorder in self._data_stream_recorders:
            recorder._resume()

    def reset_time(self):
        self._cmds_q.put(("reset_time",))

    def flush(self):
        data = {}
        for recorder in self._data_stream_recorders:
            data[recorder.name] = recorder._flush()
        return data
